from datetime import datetime, timedelta
import random
select_random = lambda x: x[int(random.random() * len(x))]

possible_values = {
    "name": ['James', 'John', 'Robert', 'Michael', 'William', 'David', 'Richard', 'Joseph', 'Thomas', 'Charles',
              'Mary', 'Patricia', 'Jennifer', 'Linda', 'Elizabeth', 'Barbara', 'Susan', 'Jessica', 'Sarah', 'Karen',
              'Emma', 'Olivia', 'Noah', 'Liam', 'Sophia', 'Ava', 'Isabella', 'Mia', 'Abigail', 'Emily',
              'Alexander', 'Ethan', 'Daniel', 'Matthew', 'Aiden', 'Henry', 'Joseph', 'Jackson', 'Samuel', 'Sebastian',
              'Sofia', 'Charlotte', 'Amelia', 'Harper', 'Evelyn', 'Aria', 'Scarlett', 'Grace', 'Chloe', 'Victoria'],
    "age": [i for i in range(10, 30)],
    "mass": [(0.5 + (0.1*i)) for i in range(10, 60)],
    "created_at": [datetime.now() - timedelta(days=i, minutes=10+20*random.random()) for i in range(10, 60)]
}

class DataGenerator:
    @staticmethod
    def name() -> str:
        return select_random(possible_values["name"])
    
    @staticmethod
    def age() -> int:
        return select_random(possible_values["age"])
    
    @staticmethod
    def mass() -> float:
        return select_random(possible_values["mass"])
    
    @staticmethod
    def created_at() -> datetime:
        return select_random(possible_values["created_at"])


from dataclasses import field, _MISSING_TYPE
from datetime import datetime
from types import NoneType
from typing import List, Union, get_origin, ClassVar, Literal, get_args, TypeVar, Any
import uuid
import polars as pl
from autodla.engine.db import DB_Connection
from autodla.engine.lambda_conversion import lambda_to_sql
from pydantic import BaseModel, GetCoreSchemaHandler, TypeAdapter
from pydantic_core import CoreSchema, core_schema, PydanticUndefinedType
from autodla.utils.logger import logger
import warnings
warnings.filterwarnings('error')

from json import JSONEncoder
def _default(self, obj):
	return getattr(obj.__class__, "to_json", _default.default)(obj)
_default.default = JSONEncoder().default
JSONEncoder.default = _default

class primary_key(str):
	@classmethod
	def generate(cls):
		return cls(str(uuid.uuid4()))
	def is_valid(self):
		try:
			uuid.UUID(self)
			return True
		except ValueError:
			return False
	@staticmethod
	def auto_increment():
		return field(default_factory=lambda: primary_key.generate())
	
	def __eq__(self, value):
		if isinstance(value, str):
			return super().__eq__(value)
		if isinstance(value, uuid.UUID):
			return uuid.UUID(self) == value
	
	@classmethod
	def __get_pydantic_core_schema__(cls, source_type: Any, handler: GetCoreSchemaHandler) -> CoreSchema:
		return core_schema.no_info_after_validator_function(cls, handler(str))

	def __hash__(self):
		return super().__hash__()

def dla_dict(operation : Literal["INSERT", "UPDATE", "DELETE"], modified_at=datetime.now(), modified_by="SYSTEM", is_current=False, is_active=True):
	def out():
		return {
			'DLA_object_id': primary_key.generate(),
			'DLA_modified_at': modified_at,
			'DLA_operation': operation,
			'DLA_modified_by': modified_by,
			'DLA_is_current': is_current,
			'DLA_is_active': is_active
		}
	return out

class Table:
	def __init__(self, table_name : str, schema : dict, db : DB_Connection = None):
		table_name_res = db.get_table_name(table_name)
		self.table_name = table_name
		self.__table_alias = table_name_res.alias
		self.schema = schema
		if db:
			self.set_db(db)
	
	@property
	def db(self) -> DB_Connection:
		db = self.__db
		if db is None:
			raise ValueError("DB not defined")
		return db
	
	def set_db(self, db : DB_Connection):
		if db is None:
			raise ValueError("DB not defined")
		self.__db = db
		self.__db.ensure_table(self.table_name, self.schema, save=True)
	
	def get_all(self, limit=10, only_current=True, only_active=True, skip=0):
		conditions = ["TRUE"]
		if only_current:
			conditions.append("DLA_is_current = true")
		if only_active:
			conditions.append("DLA_is_active = true")
		where_st = " AND ".join(conditions)
		qry = self.db.query.select(
			from_table=f'{self.__db.get_table_name(self.table_name).name} {self.__table_alias}',
			columns=[f'{self.__table_alias}.{i}' for i in list(self.schema.keys())],
			where=where_st,
			limit=limit,
			offset=skip
		)
		return self.db.execute(qry)

	def filter(self, l_func, limit=10, only_current=True, only_active=True, skip=0):
		conditions = [lambda_to_sql(self.schema, l_func, self.__db.data_transformer, alias=self.__table_alias)]
		if only_current:
			conditions.append("DLA_is_current = true")
		if only_active:
			conditions.append("DLA_is_active = true")
		where_st = " AND ".join(conditions)
		qry = self.db.query.select(
			from_table=f'{self.__db.get_table_name(self.table_name).name} {self.__table_alias}',
			columns=[f'{self.__table_alias}.{i}' for i in list(self.schema.keys())],
			where=where_st,
			limit=limit,
			offset=skip
		)
		return self.db.execute(qry)
	
	def insert(self, data : dict):
		qry = self.db.query.insert(self.__db.get_table_name(self.table_name).name, [data])
		self.db.execute(qry)
	
	def update(self, l_func, data):
		where_st = lambda_to_sql(self.schema, l_func, self.__db.data_transformer, alias="")
		update_data = {f'{key}': value for key, value in data.items()}
		qry = self.db.query.update(
			f'{self.__db.get_table_name(self.table_name).name}',
			where=where_st,
			values=update_data
		)
		return self.db.execute(qry)
	
	def delete_all(self):
		qry = self.db.query.delete(self.__db.get_table_name(self.table_name).name, "TRUE")
		self.db.execute(qry)

class Object(BaseModel):
	__table : ClassVar[Table] = None
	__dependencies : ClassVar[list] = []
	identifier_field : ClassVar[str] = "id"
	__objects_list : ClassVar[List] = []
	__objects_map : ClassVar[dict] = {}

	@classmethod
	def delete_all(cls):
		cls.__objects_list = []
		cls.__objects_map = {}
		cls.__table.delete_all()

	@classmethod
	def set_db(cls, db : DB_Connection):
		schema = cls.get_types()
		dependencies = {}
		common_fields = {
			'DLA_object_id': {
				"type": uuid.UUID
			},
			'DLA_modified_at': {
				"type": datetime
			},
			'DLA_operation': {
				"type": str
			},
			'DLA_modified_by': {
				"type": str
			},
			'DLA_is_current': {
				"type": bool
			},
			'DLA_is_active': {
				'type': bool
			}
		}
		for k, i in schema.items():
			if 'depends' in i:
				table_name = f"{cls.__name__.lower()}__{k}__{i['depends'].__name__.lower()}"
				dependencies[k] = {
					'is_list': i.get("is_list") == True,
					'is_value': False,
					'type': i['depends'],
					'table': Table(
						table_name,
						{
							"connection_id": {
								"type": primary_key
							},
							"first_id": {
								"type": primary_key
							},
							"second_id": {
								"type": primary_key
							},
							"list_index": {
								"type": int
							}
							,**common_fields
						},
						db
					)
				}
			elif 'is_list' in i:
				table_name = f"{cls.__name__.lower()}__{k}"
				dependencies[k] = {
					'is_list': i.get("is_list") == True,
					'is_value': True,
					'type': i["type"],
					'table': Table(
						table_name,
						{
							"connection_id": {
								"type": primary_key
							},
							"first_id": {
								"type": primary_key
							},
							"value": {
								"type": i["type"]
							},
							"list_index": {
								"type": int
							}
							,**common_fields
						},
						db
					)
				}
		for i in dependencies:
			del schema[i]
		cls.__table = Table(cls.__name__.lower(), {**schema,**common_fields}, db)
		cls.__dependencies = dependencies

	@classmethod
	def get_types(cls):
		out = {}
		fields = cls.model_fields
		for i in fields:
			if(get_origin(fields[i].annotation) == ClassVar):
				continue
			type_out = {}
			tp = fields[i].annotation
			ori, arg = get_origin(tp), get_args(tp)
			if ori == Union:
				if arg[1] == NoneType:
					type_out["nullable"] = True
					tp = arg[0]
					ori, arg = get_origin(tp), get_args(tp)
			if type_out.get('nullable') == True and fields[i].default != None:
				raise TypeError('Field with type Optional must initialize to None')
			if type_out.get('nullable') != True and fields[i].default == None:
				raise TypeError('Field initialized to None must be of type Optional')
			if type(fields[i].default) not in [_MISSING_TYPE, PydanticUndefinedType]:
				type_out["default"] = fields[i].default
			if type(fields[i].default_factory) not in [_MISSING_TYPE, PydanticUndefinedType]:
				type_out["default_factory"] = fields[i].default_factory
			if ori == list:
				tp = arg[0]
				ori, arg = get_origin(tp), get_args(tp)
				type_out["is_list"] = True
			if issubclass(tp, Object):
				type_out["depends"] = tp
			type_out["type"] = tp
			out[i] = type_out
		return out
	
	@classmethod
	def __update_individual(cls, data_inp):
		logger.debug(f"UPDATE INDIVIDUAL: {cls} {data_inp}")
		data = {}
		for k, v in data_inp.items():
			if not k.upper().startswith("DLA_"):
				data[k] = v
		found = cls.__objects_map.get(data[cls.identifier_field])
		try:
			cls.model_validate(data)
		except Exception as e:
			logger.error(f"Validation error for {cls.__name__} with data {data}: {e}")
			return None
		if found is not None:
			found.__dict__.update(data)
			return found
		obj = cls(**data)
		cls.__objects_list.append(obj)
		cls.__objects_map[obj[cls.identifier_field]] = obj
		return obj
	
	@classmethod
	def __update_info(cls, filter = None, limit=10, skip=0, only_current=True, only_active=True):
		if filter is None:
			res = cls.__table.get_all(limit, only_current, only_active, skip=skip)
		else:
			res = cls.__table.filter(filter, limit, only_current, only_active, skip=skip)
		obj_lis = res.to_dicts()
		if obj_lis == []:
			return []
		id_list = res[cls.identifier_field].to_list()
		
		table_results = {}
		dep_tables_required_ids = {}
		for k, v in cls.__dependencies.items():
			if v['is_value']:
				continue
			table_results[k] = v['table'].filter(lambda x: x.first_id in id_list, None, only_current=only_current, only_active=only_active)
			ids = set(table_results[k]['second_id'].to_list())
			t_name = v['type'].__name__
			if t_name not in dep_tables_required_ids:
				dep_tables_required_ids[t_name] = {"type": v['type'], "ids": ids}
			else:
				dep_tables_required_ids[t_name]["ids"] = dep_tables_required_ids[t_name]["ids"].union(ids)
		
		dep_tables = {}
		for k, v in dep_tables_required_ids.items():
			l = list(v['ids'])
			id_field = v['type'].identifier_field
			dep_tables[k] = {}
			if len(l) == 0:
				continue
			res = v['type'].filter(lambda x: x[id_field] in l)
			for obj in res:
				dep_tables[k][getattr(obj, v['type'].identifier_field)] = obj

		out = []
		for obj in obj_lis:
			for key in cls.__dependencies:
				if cls.__dependencies[key]['is_value']:
					obj_id = obj[cls.identifier_field]
					res = cls.__dependencies[key]['table'].filter(lambda x: x.first_id == obj_id)['value'].to_list()
					obj[key] = res
					continue
				df = table_results[key]
				val_lis = []
				t_name = cls.__dependencies[key]["type"].__name__
				if len(df) > 0:
					lis = df.filter(df['first_id'] == obj[cls.identifier_field])['second_id'].to_list()
					for row in lis:
						val = dep_tables[t_name].get(row)
						if val is not None:
							val_lis.append(val)
				obj[key] = val_lis
				if not cls.__dependencies[key]['is_list']:
					if obj[key] != []:
						obj[key] = obj[key][0]
					else:
						obj[key] = None
			updt = cls.__update_individual(obj)
			if updt is not None:
				out.append(updt)
		return out

	@classmethod
	def new(cls, **kwargs):
		if cls.__table is None:
			raise ImportError('DB not defined')
		if cls.identifier_field in kwargs:
			del kwargs[cls.identifier_field]
		out = cls(**kwargs)
		data = out.to_dict()
		for i in cls.__dependencies:
			del data[i]
		dla_data = dla_dict("INSERT", is_current=True)
		cls.__table.insert({**data, **dla_data()})
		for field, v in cls.__dependencies.items():
			if v['is_list']:
				new_rows = []
				if v['is_value']:
					for idx, i in enumerate(getattr(out, field)):
						new_rows.append({
							'connection_id': primary_key.generate(),
							"first_id": out[cls.identifier_field],
							"value": i,
							"list_index": idx,
							**dla_data()
						})
				else:
					for idx, i in enumerate(getattr(out, field)):
						new_rows.append({
							'connection_id': primary_key.generate(),
							"first_id": out[cls.identifier_field],
							"second_id": i[v['type'].identifier_field],
							"list_index": idx,
							**dla_data()
						})
				for j in new_rows:
					v['table'].insert(j)
			else:
				val = getattr(out, field)
				if val is not None:
					v['table'].insert({
						'connection_id': primary_key.generate(),
						"first_id": out[cls.identifier_field],
						"second_id": val[v['type'].identifier_field],
						"list_index": 0,
						**dla_data()
					})
		cls.__objects_map[str(out[cls.identifier_field])] = out
		cls.__objects_list.append(out)
		return out
	
	def history(self):
		self_res = self.__table.filter(lambda x: x[self.identifier_field] == getattr(self, self.identifier_field), limit=None, only_active=False, only_current=False)
		out = {
			"self": self_res.to_dicts(),
			"dependencies": {}
		}
		for k, v in self.__dependencies.items():
			dep_res = v['table'].filter(lambda x: x.first_id == getattr(self, self.identifier_field), limit=None, only_active=False, only_current=False)
			out['dependencies'][k] = dep_res.to_dicts()
		return out
	
	def update(self, **kwargs):
		data = {}
		for key in self.to_dict():
			if key in kwargs:
				data[key] = kwargs[key]
			elif key not in self.__class__.__dependencies:
				data[key] = getattr(self, key)
		dla_data_insert = dla_dict("UPDATE", is_current=True)
		for key, value in kwargs.items():
			if key in self.__dependencies:
				del data[key]
				dependency = self.__dependencies[key]
				dependency['table'].update(lambda x: x.first_id == self.id, {'DLA_is_current': False})
				new_rows = []
				if dependency['is_list']:
					if dependency['is_value']:
						for idx, i in enumerate(value):
							new_rows.append({
								'connection_id': primary_key.generate(),
								"first_id": self[self.identifier_field],
								"value": i,
								"list_index": idx,
								**dla_data_insert()
							})
					else:
						for idx, i in enumerate(value):
							new_rows.append({
								'connection_id': primary_key.generate(),
								"first_id": self[self.identifier_field],
								"second_id": i[dependency['type'].identifier_field],
								"list_index": idx,
								**dla_data_insert()
							})
				else:
					if value is not None:
						new_rows.append({
							'connection_id': primary_key.generate(),
							"first_id": self[self.identifier_field],
							"second_id": value[dependency['type'].identifier_field],
							"list_index": 0,
							**dla_data_insert()
						})
				for j in new_rows:
					dependency['table'].insert(j)
			setattr(self, key, value)
		self.__table.update(lambda x: x[self.identifier_field] == self.id, {'DLA_is_current': False})
		self.__table.insert({**data, **dla_data_insert()})
	
	def delete(self):
		data = {}
		for key in self.__class__.model_fields:
			data[key] = getattr(self, key)
		dla_data_delete = dla_dict("DELETE", is_current=True, is_active=False)
		for key, dependency in self.__dependencies.items():
			del data[key]
			dependency['table'].update(lambda x: x.first_id == self.id, {'DLA_is_current': False})
			value = getattr(self, key)
			if value is None:
				continue
			new_rows = []
			if dependency['is_list']:
				if dependency['is_value']:
					for idx, i in enumerate(value):
						new_rows.append({
							'connection_id': primary_key.generate(),
							"first_id": self[self.identifier_field],
							"value": i,
							"list_index": idx,
							**dla_data_delete()
						})
				else:
					for idx, i in enumerate(value):
						new_rows.append({
							'connection_id': primary_key.generate(),
							"first_id": self[self.identifier_field],
							"second_id": i[dependency['type'].identifier_field],
							"list_index": idx,
							**dla_data_delete()
						})
			else:
				if value is not None:
					new_rows.append({
						'connection_id': primary_key.generate(),
						"first_id": self[self.identifier_field],
						"second_id": value[dependency['type'].identifier_field],
						"list_index": 0,
						**dla_data_delete()
					})
			for j in new_rows:
				dependency['table'].insert(j)
		self.__table.update(lambda x: x[self.identifier_field] == self.id, {'DLA_is_current': False})
		self.__table.insert({**data, **dla_data_delete()})



	@classmethod
	def all(cls, limit=10, skip=0):
		out = cls.__update_info(limit=limit, skip=skip)
		return out

	@classmethod
	def filter(cls, lambda_f, limit=10, skip=0):
		out = cls.__update_info(filter=lambda_f, limit=limit, skip=skip)
		return out
	
	@classmethod
	def get_by_id(cls, id_param):
		cls.__update_info(lambda x: x[cls.identifier_field] == id_param, limit=1, skip=0)
		return cls.__objects_map.get(id_param)
	
	@classmethod
	def get_table_res(cls, limit=10, skip=0, only_current=True, only_active=True) -> pl.DataFrame:
		return cls.__table.get_all(limit=limit, only_current=only_current, only_active=only_active, skip=skip)
	
	def to_dict(self):
		return self.model_dump()
	
	def to_json(self):
		return self.model_dump_json()
  
	def __getitem__(self, item):
		return getattr(self, item)



import ast
import inspect
from dataclasses import dataclass
from typing import Any, Callable
from autodla.engine.data_conversion import DataTransformer, MethodArgument
from datetime import datetime
from autodla.utils.logger import logger


class LambdaFinder(ast.NodeVisitor):
    def __init__(self): 
        self.found = None

    def find(self, tree):
        self.visit(tree)
        return self.found

    def generic_visit(self, node):
        if isinstance(node, ast.Lambda):
            self.found = node
            return
        ast.NodeVisitor.generic_visit(self, node)

def lambda_to_ast(lambda_func):
    try:
        if type(lambda_func) == str:
            source = lambda_func
        else:
            source = inspect.getsource(lambda_func).strip()
    except (IOError, TypeError) as e:
        raise TypeError(f"Failed to get lambda source: {str(e)}")
    root_tree = ast.parse(source)
    return LambdaFinder().find(root_tree)

def lambda_to_text(lambda_func):
    lambda_node = lambda_to_ast(lambda_func)
    return ast.unparse(lambda_node)

@dataclass
class NodeReturn:
    st: str
    tp: type
    eval: Any = None

def pn(node):
        logger.debug("")
        if isinstance(node, ast.AST):
            logger.debug(ast.dump(node))
        else:
            logger.debug(node)
        logger.debug("")

class LambdaToSql(ast.NodeVisitor):
    def __init__(self, root : ast.Lambda, schema, data_transformer : DataTransformer, ctx_vars = {}, alias='x'):
        self.ctx_vars = ctx_vars
        self.schema = schema
        self.root = root
        self.data_transformer = data_transformer
        self.alias = alias
    
    def evaluate_node(self, node):
        return eval(str(ast.unparse(node)), self.ctx_vars)
    
    def repr(self, obj):
        st = repr(obj)
        if isinstance(obj, Callable):
            st = obj.__name__
            if hasattr(obj, '__self__'):
                st = self.repr(obj.__self) + '.' + st
        return st
    
    def obj_to_node(self, obj):
        st = self.repr(obj)
        if type(obj) in [int, float, bool, str, datetime]:
            return ast.Constant(value=obj)
        return ast.parse(st).body[0].value

    def evaluate_and_parse_node(self, node):
        out = self.obj_to_node(self.evaluate_node(node))
        return self.parse_node(out)
    
    def node_compatibility(self, node1 : NodeReturn, node2: NodeReturn):
        if isinstance(node1, type(node2)) or isinstance(node2, type(node1)):
            return True
        if node1 is None or node2 is None:
            return True
        if isinstance(node1, list):
            if all([self.node_compatibility(arg, node2) for arg in node1]):
                return True
        elif isinstance(node2, list):
            if all([self.node_compatibility(arg, node1) for arg in node2]):
                return True
        return False

    def parse_node(self, node):
        match type(node).__name__:
            case 'Subscript':
                caller = self.parse_node(node.value)
                slice_node = self.parse_node(node.slice)
                attr = slice_node.st if slice_node.eval is None else slice_node.eval
                if caller.eval is None:
                    if attr not in self.schema:
                        raise AttributeError(f"invalid attribute for {caller.st}: '{node.attr}'")
                    if caller.st == "":
                        return NodeReturn(attr, self.schema.get(attr))
                    return NodeReturn(f'{caller.st}.{attr}', self.schema.get(attr))
                if slice_node.eval is None:
                    raise ValueError(f'invalid slice node: {slice_node.st}')
                val = getattr(caller.eval, slice_node.eval)
                if val is None:
                    raise AttributeError(f"attribute not found: '{node.attr}'")
                if type(val) == Callable:
                    return NodeReturn(node.attr, Callable, val)
                return self.evaluate_and_parse_node(node)
            case 'IfExp':
                condition = self.parse_node(node.test)
                condition_value = self.parse_node(node.body)
                else_value = self.parse_node(node.orelse)
                if all([getattr(arg, 'eval', None) is not None for arg in [condition, condition_value, else_value]]):
                    return self.evaluate_and_parse_node(node)
                if condition.tp != bool:
                    raise TypeError('condition is not a valid bool')
                if not self.node_compatibility(condition_value, else_value):
                    raise TypeError('values for if should be of same type')
                st = f'(CASE WHEN {condition.st} THEN {condition_value.st} ELSE {else_value.st} END)'
                return NodeReturn(st, condition_value.tp)
            case 'Call':
                args = [self.parse_node(arg) for arg in node.args]
                caller = None
                func_name = ""
                func = None
                if type(node.func).__name__ == "Attribute":
                    caller = self.parse_node(node.func.value)
                    func_name = node.func.attr
                    if caller.eval is not None:
                        func = getattr(caller.eval, func_name)
                else:
                    func_name = node.func.id
                    f = self.ctx_vars.get(func_name)
                    if f is not None:
                        if isinstance(f, Callable):
                            func = f
                        if type(f) == type(int) and all([arg.eval is not None for arg in args]):
                            res = f(*[arg.eval for arg in args])
                            n = self.obj_to_node(res)
                            return self.parse_node(self.obj_to_node(res))
                            
                
                if func is not None and all([arg.eval is not None for arg in args]):
                    return self.evaluate_and_parse_node(node)
                
                if caller is None:
                    found = self.data_transformer.get_method(None, func_name)
                    if found is not None:
                        st, tp = found(*[MethodArgument(arg.st, arg.tp) for arg in args])
                        return NodeReturn(st, tp)
                else:
                    found = self.data_transformer.get_method(caller.tp, func_name)
                    if found is not None:
                        st, tp = found(MethodArgument(caller.st, caller.tp), *[MethodArgument(arg.st, arg.tp) for arg in args])
                        return NodeReturn(st, tp)
                raise ValueError(f'function {func_name} not found')
            
            case 'List':
                values = [self.parse_node(i) for i in node.elts]
                st_out = f'({", ".join([node_i.st for node_i in values])})'
                eval_list = [node_i.eval for node_i in values]
                if all([node_i is not None for node_i in eval_list]):
                    return NodeReturn(st_out, list, eval_list)
                return NodeReturn(st_out, list)
            case 'UnaryOp':
                op = self.data_transformer.get_operator("unary", node.op)
                child = self.parse_node(node.operand)
                if child.eval is not None:
                    return self.evaluate_and_parse_node(node)
                return NodeReturn(
                    f'({op} {child.st})',
                    bool
                )
            case 'BoolOp':
                values = [self.parse_node(i) for i in node.values]
                for i in values:
                    if i.tp != bool:
                        raise TypeError(f"expression '{i.st}' is not a valid Boolean")
                op = self.data_transformer.get_operator("boolean", node.op)
                out = f' {op} '.join([i.st for i in values])
                if all([arg.eval is not None for arg in values]):
                    return self.evaluate_and_parse_node(node)
                return NodeReturn(
                    f'({out})',
                    bool
                )
            case 'BinOp':
                left = self.parse_node(node.left)
                right = self.parse_node(node.right)
                func = self.data_transformer.get_operator("binary", node.op)
                if left.tp != right.tp:
                    raise TypeError(f"invalid operation between {left.tp} and {right.tp}")
                if left.eval is not None and right.eval is not None:
                    return self.evaluate_and_parse_node(node)
                return NodeReturn(
                    f'({func(left.st, right.st)})',
                    left.tp
                )
            case 'Name':
                if node.id == 'x':
                    return NodeReturn(self.alias, str)
                value = self.ctx_vars.get(node.id)
                if value is not None:
                    try:
                        transformed_value = self.data_transformer.convert_data(value)
                    except:
                        transformed_value = node.id
                    return NodeReturn(transformed_value, type(value), value)
                raise ValueError(f"'{node.id}' is not defined")
            case 'Constant':
                return NodeReturn(
                    self.data_transformer.convert_data(node.value),
                    type(node.value),
                    node.value
                )
            case 'Compare':
                left = self.parse_node(node.left)
                op = self.data_transformer.get_operator("numeric", node.ops[0])
                right = self.parse_node(node.comparators[0])
                if not self.node_compatibility(left, right):
                    raise TypeError(f"invalid comparission between {left.tp} and {right.tp}")
                if left.eval is not None and right.eval is not None:
                    return self.evaluate_and_parse_node(node)
                return NodeReturn(
                    f'({left.st} {op} {right.st})', bool
                )
            case 'Attribute':
                caller = self.parse_node(node.value)
                if caller.eval is None:
                    if node.attr not in self.schema:
                        raise AttributeError(f"invalid attribute for x: '{node.attr}'")
                    if caller.st == "":
                        return NodeReturn(node.attr, self.schema.get(node.attr))
                    return NodeReturn(f'{caller.st}.{node.attr}', self.schema.get(node.attr))
                val = getattr(caller.eval, node.attr)
                if val is None:
                    raise AttributeError(f"attribute not found: '{node.attr}'")
                if type(val) == Callable:
                    return NodeReturn(node.attr, Callable, val)
                return self.evaluate_and_parse_node(node)
            case _:
                try:
                    evaluation = self.evaluate_node(node)
                    new_node=ast.Constant(value=evaluation)
                    return self.parse_node(new_node)
                except:
                    logger.debug(ast.dump(node))
                    raise ValueError(f'Invalid node type: {type(node).__name__}')


    def transform(self):
        definition_error = False
        if len(self.root.args.args) != 1:
            definition_error = True
        if list(self.root.args.args)[0].arg != 'x':
            definition_error = True
        if definition_error:
            raise SyntaxError("lambda definition should be -> 'lambda x:'")
        out = self.parse_node(self.root.body)
        return out

import builtins
def get_context_from_lamba(lambda_func):
    file = lambda_func.__code__.co_filename
    line = lambda_func.__code__.co_firstlineno
    found = None
    for frame in inspect.stack():
        if frame.filename == file and frame.lineno == line:
            found = frame
            break
    if found is None:
        raise ValueError('frame not found')
    return {**vars(builtins), **found.frame.f_locals}

def lambda_to_sql(schema, lambda_func, data_transformer : DataTransformer, ctx_vars={}, alias='x') -> str:
    if not type(lambda_func) == str:
        ctx_vars = get_context_from_lamba(lambda_func)
    lambda_node = lambda_to_ast(lambda_func)
    out = LambdaToSql(lambda_node, schema, data_transformer=data_transformer, ctx_vars=ctx_vars, alias=alias).transform()
    return out.st

def json_to_lambda_str(json_condition):
    """
    Transforms a SQL-inspired JSON condition to a Python lambda string representation.
    
    Args:
        json_condition (dict): A condition object that can be:
            - Simple: {"field": "age", "operator": "gt", "value": 10}
            - Complex: {"and": [condition1, condition2, ...]} or {"or": [condition1, condition2, ...]}
    
    Returns:
        str: A string representation of the lambda function
    """
    # Check if this is a complex condition with AND/OR
    if "and" in json_condition:
        sub_conditions = [json_to_lambda_str(cond) for cond in json_condition["and"]]
        return f"lambda x: {' and '.join(f'({cond})' for cond in sub_conditions)}"
    
    elif "or" in json_condition:
        sub_conditions = [json_to_lambda_str(cond) for cond in json_condition["or"]]
        return f"lambda x: {' or '.join(f'({cond})' for cond in sub_conditions)}"
    
    # Handle negation
    elif "not" in json_condition:
        sub_condition = json_to_lambda_str(json_condition["not"])
        # Extract the condition part (after "lambda x: ")
        cond_part = sub_condition.split("lambda x: ", 1)[1]
        return f"lambda x: not ({cond_part})"
    
    # Handle simple condition
    elif all(k in json_condition for k in ["field", "operator"]):
        field = json_condition.get("field")
        operator = json_condition.get("operator")
        value = json_condition.get("value")
        
        # Map of operators to Python comparison operators
        operator_map = {
            "eq": "==",
            "neq": "!=",
            "gt": ">",
            "gte": ">=",
            "lt": "<",
            "lte": "<=",
            "in": "in",
            "nin": "not in"
        }
        
        if operator not in operator_map:
            raise ValueError(f"Unsupported operator: {operator}")
        
        # Generate the lambda function string
        op_str = operator_map[operator]
        
        # Format the value appropriately
        if isinstance(value, str):
            formatted_value = f"'{value}'"
        elif isinstance(value, list):
            # Format each element in the list
            formatted_elements = []
            for elem in value:
                if isinstance(elem, str):
                    formatted_elements.append(f"'{elem}'")
                else:
                    formatted_elements.append(str(elem))
            formatted_value = f"[{', '.join(formatted_elements)}]"
        else:
            formatted_value = str(value)
        
        return f"lambda x: x.{field} {op_str} {formatted_value}"
    else:
        raise ValueError(f"Invalid condition format: {json_condition}")


import polars as pl
from typing import Callable, List, Optional
from autodla.engine.data_conversion import DataTransformer

class QueryBuilder:
    def __init__(self, data_transformer = DataTransformer):
        self._data_transformer = data_transformer

    def select(self, from_table: str, columns: List[str], where: str = None, limit: int = 10, order_by: str = None, group_by: list[str] = None, offset: int = None) -> pl.DataFrame:
        pass

    def insert(self, into_table: str, values: List[dict]) -> None:
        pass

    def update(self, table: str, values: dict, where: str) -> None:
        pass

    def delete(self, table: str, where: str) -> None:
        pass

    def create_table(self, table_name: str, schema: dict, if_exists = False) -> None:
        pass

    def drop_table(self, table_name: str, if_exists = False) -> None:
        pass


import os
from functools import wraps
from typing import Callable
import inspect

import hashlib
def hash(st):
    return hashlib.sha256(st.encode()).hexdigest()

AUTODLAWEB_USER = 'autodla'
if "AUTODLAWEB_USER" in os.environ:
    AUTODLAWEB_USER = os.environ.get("AUTODLAWEB_USER")
AUTODLAWEB_PASSWORD = hash('password')
if "AUTODLAWEB_PASSWORD" in os.environ:
    AUTODLAWEB_PASSWORD = hash(os.environ.get("AUTODLAWEB_PASSWORD"))


import uuid
def generate_token():
    return hash(str(uuid.uuid4()))

class EndpointMaker:
    @classmethod
    def list(cls, object) -> Callable:
        pass

    @classmethod
    def get(cls, object) -> Callable:
        pass

    @classmethod
    def get_history(cls, object) -> Callable:
        pass

    @classmethod
    def table(cls, object) -> Callable:
        pass

    @classmethod
    def new(cls, object) -> Callable:
        pass

    @classmethod
    def edit(cls, object) -> Callable:
        pass

    @classmethod
    def delete(cls, object) -> Callable:
        pass


class WebConnection:
    def __init__(self, endpoint_maker: EndpointMaker, setup_autodla_web=True):
        self.endpoint_maker = endpoint_maker
        self.current_token = generate_token()
        self.setup_admin_endpoints()
        if setup_autodla_web:
            self.setup_autodla_web_static_files()
            self.setup_autodla_web_endpoints()

    def setup_admin_endpoints(self):
        pass

    def setup_autodla_web_static_files(self):
        import os
        from importlib import resources as impresources
        from .. import static as staticdir
        import tempfile
        import shutil
        temp_dir = tempfile.mkdtemp()
        static_package_dir = impresources.files(staticdir)
        static_temp_dir = os.path.join(temp_dir, 'static')
        os.makedirs(static_temp_dir, exist_ok=True)
        def copy_dir_recursively(source_dir, dest_dir):
            os.makedirs(dest_dir, exist_ok=True)
            for item in source_dir.iterdir():
                dest_path = os.path.join(dest_dir, item.name)
                if item.is_file():
                    with item.open('rb') as src, open(dest_path, 'wb') as dst:
                        shutil.copyfileobj(src, dst)
                elif item.is_dir():
                    copy_dir_recursively(item, dest_path)
        copy_dir_recursively(static_package_dir, static_temp_dir)
        self.static_temp_dir = static_temp_dir

    def setup_autodla_web_endpoints(self):
        pass
    
    ###### Admin token management
    
    _current_admin_token = ""
    
    def create_new_admin_token(self):
        new_token = generate_token()
        self._current_admin_token = new_token
        return new_token
    
    def validate_token(self, token):
        if token != self._current_admin_token:
            self.unauthorized_handler()
        return True
    
    def login(self, username, password):
        if username != AUTODLAWEB_USER or hash(password) != AUTODLAWEB_PASSWORD:
            self.invalid_admin_credentials_handler()
        return self.create_new_admin_token()
    
    def normalize_endpoint(self, func):
        return func

    def admin_endpoint_validate(self, func):
        return func
    
    async def extract_token(self, *args, **kwargs):
        pass

    def admin_endpoint(self, func):
        self.admin_endpoint_validate(func)
        func = self.normalize_endpoint(func)
        @wraps(func)
        async def wrapper(*args, **kwargs):
            token = await self.extract_token(*args, **kwargs)
            self.validate_token(token)
            return await func(*args, **kwargs)
        return wrapper
    
    ### Handlers
    @classmethod
    def unauthorized_handler(cls):
        raise NotImplementedError("This method should be implemented in subclasses.")
    @classmethod
    def invalid_admin_credentials_handler(cls):
        raise NotImplementedError("This method should be implemented in subclasses.")




import polars as pl
from pydantic import BaseModel
from autodla.engine.data_conversion import DataTransformer
from autodla.engine.query_builder import QueryBuilder
from typing import get_origin, get_args
from autodla.utils.logger import logger


class TableName(BaseModel):
    name: str
    alias: str

class DB_Connection:
    __data_transformer : DataTransformer
    __query : QueryBuilder

    @property
    def usage_metrics(self):
        """
        Returns the usage metrics of the database connection.
        This is a placeholder method and should be implemented in subclasses.
        """
        raise NotImplementedError("This method should be implemented in subclasses.")

    def __init__(self, data_transformer, query):
        self.__data_transformer = data_transformer
        self.__query = query
        self.__classes = {}
        self._table_schemas = {}
        self._tables = []

    @property
    def query(self):
        return self.__query

    @property
    def data_transformer(self):
        return self.__data_transformer
    
    def get_table_name(self, table_name: str) -> TableName:
        """
        Returns the table name in uppercase.
        This is used to ensure that the table names are consistent with the SQL standard.
        """
        return TableName(name=table_name.upper(), alias=table_name.lower())
    
    def clean_db(self, DO_NOT_ASK=False):
        if not DO_NOT_ASK:
            logger.debug("Are you sure you want to clean the database? (y/n)")
            answer = input()
            if answer != "y":
                raise Exception("User did not confirm the action")
        logger.debug("Cleaning database...")
        for class_i in self.__classes.values():
            class_i.delete_all()
        logger.debug("Database cleaned")

    
    def get_table_definition(self, table_name) -> dict[str, type]:
        pass
    
    def attach(self, objects):
        ordered_objects = []
        pending = objects[:]
        while True:
            if pending == []:
                break
            tmp = pending[:]
            for obj in tmp:
                schema = obj.get_types()
                class_dependencies = []
                for i in schema.values():
                    if 'depends' in i:
                        class_dependencies.append(i.get('depends'))
                if all([dep in ordered_objects for dep in class_dependencies]):
                    ordered_objects.append(obj)
                    pending.remove(obj)
        for obj in ordered_objects:
            self.__classes[obj.__name__] = obj
            obj.set_db(self)
    
    def get_json_schema(self):
        out = {}
        for class_key, class_i in self.__classes.items():
            class_def = class_i.get_types()
            class_out = {}
            for k, f in class_def.items():
                class_out[k] = {}
                if class_i.identifier_field == k:
                    class_out[k]["primary_key"] = True
                if "depends" in f:
                    class_out[k]["depends"] = f'$ref:{f["depends"].__name__}'
                if "is_list" in f:
                    class_out[k]["is_list"] = f["is_list"]
                if "nullable" in f:
                    class_out[k]["nullable"] = f["nullable"]
                type_st = f["type"].__name__
                if get_origin(f['type']) == list:
                    arg = get_args(f["type"])
                    if len(arg) == 1:
                        type_st += f'[{arg[0].__name__}]'
                class_out[k]["type"] = type_st
            out[class_key] = class_out
        return out
    
    @property
    def classes(self):
        return self.__classes.values()

    def execute(self, query: str) -> pl.DataFrame:
        pass

    def normalize_statment(self, statement: str) -> str:
        if not isinstance(statement, str):
            statement = str(statement)
        statement = statement.lstrip().rstrip()
        if statement[-1] != ";":
            statement += ";"
        return statement
    
    def ensure_table(self, table_name, schema, save=False, current_data_schema=None):
        logger.debug("ENSURE TABLE DBCONNECTION")
        if save:
            self._table_schemas[table_name] = schema
            self._tables.append(table_name)
        data_schema = {k.upper(): v["type"] for k, v in schema.items()}
        if current_data_schema is None:
            current_data_schema = self.get_table_definition(table_name)
        if all([self.data_transformer.check_type_compatibilty(data_schema.get(k), current_data_schema.get(k)) for k in list(set(data_schema.keys()).union(set(data_schema.keys())))]):
            return
        logger.debug(data_schema)
        logger.debug(current_data_schema)
        if data_schema == current_data_schema:
            return
        schema = self.data_transformer.convert_data_schema(schema)
        table_name_db = self.get_table_name(table_name)
        self.execute(self.query.drop_table(table_name_db.name, if_exists=True))
        qry = self.query.create_table(table_name_db.name, schema)
        self.execute(qry)


from datetime import datetime, date
from dataclasses import dataclass
from types import NoneType
from typing import Callable, Any, Literal, Optional, Union, get_args, get_origin
import os

DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'
if "DATETIME_FORMAT" in os.environ:
    DATETIME_FORMAT = os.environ.get("DATETIME_FORMAT")

@dataclass
class DataConversion:
    name: str
    transform: Callable[[Any], str] = lambda x: f"{x}"

@dataclass
class MethodArgument:
    st: str
    tp: type

class GlobalTypedMethod:
    def __init__(self, return_type: type, definition : str, args : dict[type]):
        self.return_type = return_type
        self.definition = definition
        self.args = args
    
    def check_type(self, expected_type, compared_type):
        if type(expected_type) == list:
            for type_i in expected_type:
                if type_i is None and compared_type is None:
                    return True
                if compared_type == type_i:
                    return True
        elif compared_type == expected_type:
            return True
        return False
    
    def __call__(self, *args, **kwds):
        keys = list(self.args.keys())
        arguments = {}
        for i, k in enumerate(keys):
            v = None
            if i < len(args):
                v = args[i]
            if k in kwds:
                v = kwds[k]
            if not self.check_type(self.args[k], getattr(v, 'tp', None)):
                raise TypeError(f"expected type {self.args[k]} for argument {k}, got {type(v)}")
            arguments[k] = getattr(v, 'st', None)
        return self.definition(arguments), self.return_type

class TypedMethod(GlobalTypedMethod):
    def __init__(self, caller_type: type, return_type: type, definition : str, args : dict[type]):
        self.caller_type = caller_type
        super().__init__(return_type, definition, args)
    
    def __call__(self, caller: MethodArgument, *args, **kwds):
        if any([(type(arg) != MethodArgument) for arg in args]) or any([(type(arg) != MethodArgument) for arg in list(kwds.values())]):
            raise ValueError("All arguments for data conversion should be MethodArgument")
        keys = list(self.args.keys())
        arguments = {}
        for i, k in enumerate(keys):
            v = None
            if i < len(args):
                v = args[i]
            if k in kwds:
                v = kwds[k]
            if not self.check_type(self.args[k], v.tp if v is not None else None):
                raise TypeError(f"expected type {self.args[k]} for argument {k}, got {v.tp}")
            arguments[k] = v.st
        if not self.check_type(self.caller_type, caller.tp):
            raise TypeError(f"expected caller type {self.caller_type}, got {caller.tp}")
        arguments["caller"] = caller.st
        return self.definition(arguments), self.return_type

class DataTransformer:
    TYPE_DICT = {}
    OPERATOR_DICT = {
        "numeric": {
            'Eq': "="
        },
        "binary": {
            "Add": lambda x, y: f'{x} + {y}',
        },
        "boolean": {
            "And": 'AND',
        },
        "unary": {
            "Not": 'NOT'
        }
    }
    METHODS_MAP = {
        "global": {
            'round': GlobalTypedMethod(
                int,
                lambda data: f"ROUND({data['value']}{str(', ' + data['digits']) if data['digits'] is not None else ''})",
                {"value": [float, int], 'digits': [int, None]}
            ),
            'abs': GlobalTypedMethod(
                int,
                lambda data: f'ABS({data["value"]})',
                {'value': [float, int]}
            )
        },
        "by_type": {
            str: {
                "lower": TypedMethod(
                    str, str,
                    lambda data: f"LOWER({data['caller']})",
                    {}
                ),
                "startswith": TypedMethod(
                    str, bool,
                    lambda data: f"({data['caller']} LIKE '{data['value'][1:-1]}%')",
                    {"value": str}
                )
            },
            datetime: {
                "year": TypedMethod(
                    datetime, int,
                    lambda data: f"EXTRACT(YEAR FROM {data['caller']})",
                    {}
                ),
            }
        }
    }

    NODE_COMPATIBILITY = {}

    @classmethod
    def check_type_compatibilty(cls, tp1, tp2):
        out = tp1 == tp2
        if not out:
            if tp1 in cls.NODE_COMPATIBILITY:
                out = cls.NODE_COMPATIBILITY[tp1] == tp2
        return out

    @classmethod
    def get_type_from_sql_type(cls, sql_type) -> type:
        found = None
        for k, v in cls.TYPE_DICT.items():
            if v.name.upper() == sql_type.upper():
                found = k
        if found is None:
            raise TypeError(f"invalid conversion for sql_type '{sql_type}'")
        return found

    @classmethod
    def get_method(cls, caller_type : type, method_name : str):
        d = cls.METHODS_MAP["global"]
        if caller_type is not None:
            d = cls.METHODS_MAP["by_type"][caller_type]
        return d.get(method_name)

    @classmethod
    def get_operator(cls, operator_type: Literal["numeric", "binary", "boolean", "unary"], op):
        """Convert Python comparison operators to SQL operators"""
        out = cls.OPERATOR_DICT[operator_type].get(type(op).__name__)
        if out is None:
            raise ValueError(f"Unsupported {operator_type} operator: {op.__class__.__name__}")
        return out

    @classmethod
    def get_data_field(cls, v) -> DataConversion:
        if get_origin(v) is Union:
            a = get_args(v)
            if a[1] is NoneType:
                v = a[0]
        return cls.TYPE_DICT.get(v)

    @classmethod
    def convert_data_schema(cls, schema):
        out = {}
        for k, v in schema.items():
            f = cls.get_data_field(v["type"])
            if f is not None:
                out[k] = f.name
                if v.get("nullable") != True:
                    out[k] += " NOT NULL"
        return out
    
    @staticmethod
    def validate_data_from_schema(schema, data):
        extra_keys = []
        for i in data:
            if i not in schema:
                extra_keys.append(i)
        if extra_keys:
            raise ValueError(f"Exta values found: {extra_keys}")
        missing_values = []
        for i in data:
            if i not in schema:
                missing_values.append(i)
        if missing_values:
            raise ValueError(f"Missing values: {missing_values}")
        invalid_types = []
        for i in data:
            if not isinstance(data[i], schema[i]):
                invalid_types.append([i, type(data[i]), schema[i]])
        if invalid_types:
            msg = 'Invalid types:\n'
            for i in invalid_types:
                msg +=  f'{i}: expected {schema[i]} got {type(data[i])}\n'
            raise ValueError(msg)
    
    @classmethod
    def convert_data(cls, data):
        v = cls.get_data_field(type(data))
        if v is not None:
            return v.transform(data)
        if type(data) == list:
            return f"({', '.join([cls.convert_data(i) for i in data])})"
        raise TypeError(f"Missing transformer for class {type(data).__name__}")